package de.hft.stuttgart.citydoctor2.webservice.endpoints;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.file.Files;
import java.nio.file.StandardCopyOption;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.zip.ZipInputStream;

import javax.persistence.EntityManager;
import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.Status;
import javax.ws.rs.core.StreamingOutput;

import org.apache.commons.io.output.ByteArrayOutputStream;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.glassfish.jersey.media.multipart.FormDataContentDisposition;
import org.glassfish.jersey.media.multipart.FormDataParam;

import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.Checker;
import de.hft.stuttgart.citydoctor2.check.ValidationConfiguration;
import de.hft.stuttgart.citydoctor2.checks.util.FeatureCheckedListener;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.webservice.AuthenticationException;
import de.hft.stuttgart.citydoctor2.webservice.database.BuildingQueries;
import de.hft.stuttgart.citydoctor2.webservice.database.ErrorQueries;
import de.hft.stuttgart.citydoctor2.webservice.database.ValidationQueries;
import de.hft.stuttgart.citydoctor2.webservice.model.ValidationReference;
import de.hft.stuttgart.citydoctor2.webservice.model.ValidationStatus;
import de.hft.stuttgart.citydoctor2.webservice.model.entities.FeatureValidationStatus;
import de.hft.stuttgart.citydoctor2.webservice.model.entities.User;
import de.hft.stuttgart.citydoctor2.webservice.model.entities.Validation;
import de.hft.stuttgart.citydoctor2.webservice.model.entities.WebBuilding;
import de.hft.stuttgart.citydoctor2.webservice.model.entities.WebCityDoctorModel;
import de.hft.stuttgart.citydoctor2.webservice.model.entities.WebError;
import de.hft.stuttgart.citydoctor2.webservice.utils.AuthenticationUtils;
import de.hft.stuttgart.citydoctor2.webservice.utils.HibernateUtils;

@Path("models")
public class ValidationEndpoint {

	private static final Logger logger = LogManager.getLogger(ValidationEndpoint.class);

	private static final AtomicInteger FILE_COUNTER = new AtomicInteger(0);

	@Path("{modelId}/validations")
	@POST
	@Consumes(MediaType.MULTIPART_FORM_DATA)
	@Produces(MediaType.APPLICATION_JSON)
	public Response addValidation(@FormDataParam("file") InputStream fileStream,
			@FormDataParam("file") FormDataContentDisposition fileDetails, @Context HttpServletRequest request,
			@PathParam("modelId") int modelId, @Context ServletContext context) {
		EntityManager manager = null;
		try {
			manager = HibernateUtils.createManager();
			User user = AuthenticationUtils.authenticateUser(request, manager);
			if (user == null) {
				return Response.status(Status.FORBIDDEN).build();
			}
			WebCityDoctorModel model = manager.find(WebCityDoctorModel.class, modelId);
			if (model.getOwner().getId() != user.getId()) {
				return Response.status(Status.FORBIDDEN).build();
			}
			ValidationConfiguration config = ValidationConfiguration.loadValidationConfig(fileStream);
			Validation validation = new Validation(model, ValidationStatus.VALIDATING, fileDetails.getFileName());
			ValidationQueries.persistValidation(manager, model, validation);

			startValidationProcess(context, model, config, validation);

			ValidationReference validRef = new ValidationReference(validation.getId(), fileDetails.getFileName(),
					validation.getStatus());
			return Response.ok(validRef).build();
		} catch (AuthenticationException e) {
			logger.trace(AuthenticationUtils.AUTHENTICATION_DENIED, request.getRemoteAddr(), e);
			return Response.status(Status.FORBIDDEN).build();
		} finally {
			if (manager != null) {
				manager.close();
			}
		}
	}

	@Path("validations/{validationId}/status")
	@GET
	@Produces(MediaType.APPLICATION_JSON)
	public Response getValidationStatus(@PathParam("validationId") int validationId,
			@Context HttpServletRequest request) {
		EntityManager manager = null;
		try {
			manager = HibernateUtils.createManager();
			User user = AuthenticationUtils.authenticateUser(request, manager);
			if (user == null) {
				return Response.status(Status.FORBIDDEN).build();
			}
			manager.getTransaction().begin();
			Validation validation = manager.find(Validation.class, validationId);
			manager.getTransaction().commit();
			if (validation == null) {
				return Response.status(Status.BAD_REQUEST).build();
			}
			if (user.getId() != validation.getModel().getOwner().getId()) {
				return Response.status(Status.FORBIDDEN).build();
			}
			ValidationReference validRef = new ValidationReference(validation.getId(), validation.getName(),
					validation.getStatus());
			return Response.ok(validRef).build();
		} catch (AuthenticationException e) {
			logger.trace(AuthenticationUtils.AUTHENTICATION_DENIED, request.getRemoteAddr(), e);
			return Response.status(Status.FORBIDDEN).build();
		} finally {
			if (manager != null) {
				manager.close();
			}
		}
	}

	@Path("validations/{validationId}/xml")
	@GET
	@Produces(MediaType.APPLICATION_XML)
	public Response getXmlReport(@Context HttpServletRequest request, @PathParam("validationId") int validationId) {
		EntityManager manager = null;
		try {
			manager = HibernateUtils.createManager();
			User user = AuthenticationUtils.authenticateUser(request, manager);
			if (user == null) {
				return Response.status(Status.FORBIDDEN).build();
			}
			Validation valid = ValidationQueries.getValidation(validationId, manager);
			WebCityDoctorModel model = valid.getModel();
			if (model.getOwner().getId() != user.getId()) {
				return Response.status(Status.FORBIDDEN).build();
			}

			if (valid.getStatus() != ValidationStatus.OK) {
				Response.status(Status.BAD_REQUEST).build();
			}

			StreamingOutput out = new StreamingOutput() {

				@Override
				public void write(OutputStream output) throws IOException {
					output.write(valid.getXmlReport());
				}
			};
			int lastIndexOf = model.getFileName().lastIndexOf('.');
			String fileName = model.getFileName();
			if (lastIndexOf > 0) {
				fileName = model.getFileName().substring(0, lastIndexOf);
			}
			fileName = "report_" + fileName + ".xml";
			return Response.ok(out).header("Content-Disposition", "attachment; filename=" + fileName).build();
		} catch (AuthenticationException e) {
			logger.trace(AuthenticationUtils.AUTHENTICATION_DENIED, request.getRemoteAddr(), e);
			return Response.status(Status.FORBIDDEN).build();
		} finally {
			if (manager != null) {
				manager.close();
			}
		}
	}

	@Path("validations/{validationId}/pdf")
	@GET
	@Produces("application/pdf")
	public Response getPdfReport(@Context HttpServletRequest request, @PathParam("validationId") int validationId) {
		EntityManager manager = null;
		try {
			manager = HibernateUtils.createManager();
			User user = AuthenticationUtils.authenticateUser(request, manager);
			if (user == null) {
				return Response.status(Status.FORBIDDEN).build();
			}
			Validation valid = ValidationQueries.getValidation(validationId, manager);
			WebCityDoctorModel model = valid.getModel();
			if (model.getOwner().getId() != user.getId()) {
				return Response.status(Status.FORBIDDEN).build();
			}

			if (valid.getStatus() != ValidationStatus.OK) {
				Response.status(Status.BAD_REQUEST).build();
			}

			StreamingOutput out = new StreamingOutput() {

				@Override
				public void write(OutputStream output) throws IOException {
					output.write(valid.getPdfReport());
				}
			};
			int lastIndexOf = model.getFileName().lastIndexOf('.');
			String fileName = model.getFileName();
			if (lastIndexOf > 0) {
				fileName = model.getFileName().substring(0, lastIndexOf);
			}
			fileName = "report_" + fileName + ".pdf";
			return Response.ok(out).header("Content-Disposition", "attachment; filename=" + fileName).build();
		} catch (AuthenticationException e) {
			logger.trace(AuthenticationUtils.AUTHENTICATION_DENIED, request.getRemoteAddr(), e);
			return Response.status(Status.FORBIDDEN).build();
		} catch (Exception e) {
			logger.fatal("Error while fetching pdf report for validation id " + validationId, e);
			return Response.serverError().build();
		} finally {
			if (manager != null) {
				manager.close();
			}
		}
	}

	@Path("validations/{validationId}")
	@DELETE
	public Response deleteValidation(@Context HttpServletRequest request, @PathParam("validationId") int validationId) {
		EntityManager manager = null;
		try {
			manager = HibernateUtils.createManager();
			User user = AuthenticationUtils.authenticateUser(request, manager);
			if (user == null) {
				return Response.status(Status.FORBIDDEN).build();
			}
			Validation valid = ValidationQueries.getValidation(validationId, manager);
			WebCityDoctorModel model = valid.getModel();
			if (model.getOwner().getId() != user.getId()) {
				return Response.status(Status.FORBIDDEN).build();
			}
			manager.getTransaction().begin();
			manager.remove(valid);
			manager.getTransaction().commit();
			return Response.ok().build();
		} catch (AuthenticationException e) {
			logger.trace(AuthenticationUtils.AUTHENTICATION_DENIED, request.getRemoteAddr(), e);
			return Response.status(Status.FORBIDDEN).build();
		} catch (Exception e) {
			logger.fatal("Error while deleting validation id " + validationId, e);
			return Response.serverError().build();
		} finally {
			if (manager != null) {
				manager.close();
			}
		}
	}

	private void startValidationProcess(ServletContext context, WebCityDoctorModel oldModel,
			ValidationConfiguration config, Validation oldValidation) {
		Thread t = new Thread(() -> {
			final File tempDir = (File) context.getAttribute("javax.servlet.context.tempdir");
			String pdfFileName = FILE_COUNTER.getAndIncrement() + ".pdf";
			String tempPdfPath = tempDir.getAbsolutePath() + File.separator + pdfFileName;
			String xmlFileName = FILE_COUNTER.getAndIncrement() + ".xml";
			String tempXmlPath = tempDir.getAbsolutePath() + File.separator + xmlFileName;
			EntityManager manager = HibernateUtils.createManager();
			WebCityDoctorModel model = manager.merge(oldModel);
			Validation validation = manager.merge(oldValidation);

			File gmlFile = null;
			try {
				String gmlFileName = FILE_COUNTER.getAndIncrement() + ".gml";
				String gmlFileString = tempDir.getAbsolutePath() + File.separator + gmlFileName;
				gmlFile = new File(gmlFileString);
				unpackGmlFile(model, gmlFile);

				FeatureCheckedListener l = co -> {
					Set<CheckError> errors = new HashSet<>();
					List<CheckError> errorList = new ArrayList<>();
					co.collectContainedErrors(errorList);
					errors.addAll(errorList);
					WebBuilding building = BuildingQueries.getBuildingForGmlId(co.getGmlId().getGmlString(),
							model.getId(), manager);
					FeatureValidationStatus status = new FeatureValidationStatus();
					status.setWasValidated(co.hasCheckResults());
					status.setHasError(!errors.isEmpty());
					ValidationQueries.addFeatureValidationStatus(building, validation, status, manager);
					if (!errors.isEmpty()) {
						for (CheckError err : errors) {
							WebError webError = new WebError();
							webError.setBuilding(building);
							webError.setName(err.getErrorId().toString());
							webError.setValidation(validation);
							ErrorQueries.addErrorToBuilding(building, validation, webError, manager);
						}
					}
				};
				Checker.streamCheck(gmlFile, tempXmlPath, tempPdfPath, config, l, null);

				ValidationQueries.updateValidationWithReports(validation, readToByteArray(tempPdfPath),
						readToByteArray(tempXmlPath), manager);
				ValidationQueries.updateStatus(validation, ValidationStatus.OK, manager);
			} catch (IOException | CityGmlParseException e1) {
				logger.fatal("Error while validating {}", model.getFileName());
				logger.catching(e1);
				ValidationQueries.updateStatus(validation, ValidationStatus.ERROR, manager);
			} finally {
				manager.close();
				try {
					if (gmlFile != null && gmlFile.exists()) {
						Files.delete(gmlFile.toPath());
					}
					File pdfFile = new File(tempPdfPath);
					if (pdfFile.exists()) {
						Files.delete(pdfFile.toPath());
					}
					File xmlFile = new File(tempXmlPath);
					if (xmlFile.exists()) {
						Files.delete(xmlFile.toPath());
					}
				} catch (IOException e1) {
					logger.catching(e1);
				}

			}
		});
		t.setUncaughtExceptionHandler((thread, e) -> {
			logger.catching(e);
			setValidationStatusToError(oldValidation);
		});
		t.start();

	}

	private byte[] readToByteArray(String tempPdfPath) throws IOException {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		File pdfFile = new File(tempPdfPath);
		Files.copy(pdfFile.toPath(), baos);
		return baos.toByteArray();
	}

	private void setValidationStatusToError(Validation validation) {
		EntityManager manager = HibernateUtils.createManager();
		Validation valid = manager.merge(validation);
		ValidationQueries.updateStatus(valid, ValidationStatus.ERROR, manager);
		manager.close();
	}

	private void unpackGmlFile(WebCityDoctorModel model, File gmlFile) throws IOException {
		File zipFile = new File(model.getZipFile().getFilePath());
		try (ZipInputStream zis = new ZipInputStream(new FileInputStream(zipFile))) {
			zis.getNextEntry();
			Files.copy(zis, gmlFile.toPath(), StandardCopyOption.REPLACE_EXISTING);
		}
	}

}
